import os, stat, shutil, zipfile
from typing import List, Tuple
import time, ssl, pathlib, random
from .utils.system import System, console
from .utils.sub_process import execute
from urllib import request
from datetime import datetime
from dateutil import tz

def is_file_modified(src: str, dst: str):
    return os.path.getmtime(src) == os.path.getmtime(dst) and\
           os.path.getsize(src) == os.path.getsize(dst)


def find_executable_from_system(name: str):
    if not name: return ''
    output = shutil.which(name)
    if not output: return ''
    return output if os.path.isfile(output) else ''


def copy_files(fromFolder: str, toFolder: str, fileNames: List[str]):
    for file in fileNames:
        sourceFile = os.path.join(fromFolder, file)
        if not os.path.exists(sourceFile):
            continue
        destFile = os.path.join(toFolder, os.path.basename(file))
        destFileExists = os.path.exists(destFile)
        if destFileExists and is_file_modified(sourceFile, destFile):
            console(f"skipping copy '{destFile}'")
            continue
        console(f"copyto '{toFolder}'  '{sourceFile}'")
        if System.windows and destFileExists: # note: windows crashes if dest file is in use
            tempCopy = f'{destFile}.{random.randrange(1000)}.deleted'
            shutil.move(destFile, tempCopy)
            try:
                os.remove(tempCopy)
            except Exception:
                pass
        shutil.copy2(sourceFile, destFile) # copy while preserving metadata


def deploy_framework(framework: str, deployFolder: str):
    if not os.path.exists(framework):
        raise IOError(f'no framework found at: {framework}') 
    if os.path.exists(deployFolder):
        name = os.path.basename(framework)
        deployPath = os.path.join(deployFolder, name)
        console(f'Deploying framework to {deployPath}')
        execute(f'rm -rf {deployPath}')
        shutil.copytree(framework, deployPath)
        return True
    return False


def has_contents_changed(filename: str, new_contents: str):
    if not os.path.exists(filename):
        return True
    return read_text_from(filename) != new_contents


def save_file_if_contents_changed(filename: str, new_contents: str):
    if not has_contents_changed(filename, new_contents):
        return
    write_text_to(filename, new_contents)


def path_join(first: str, second: str) -> str:
    """ Always join with forward/ slashes """
    first  = first.rstrip('/\\')
    second = second.lstrip('/\\')
    if not first: return second
    if not second: return first
    return first + '/' + second


def forward_slashes(pathstring: str) -> str:
    """ Replace all back\\ slashes with forward/ slashes"""
    return pathstring.replace('\\', '/')


def back_slashes(pathstring: str) -> str:
    """ Replace all forward/ slashes with back\\ slashes"""
    return pathstring.replace('/', '\\')


def normalized_path(pathstring: str) -> str:
    """ Normalizes a path to ABSOLUTE path and all forward/ slashes """
    pathstring = os.path.abspath(pathstring)
    return pathstring.replace('\\', '/').rstrip()


def normalized_join(path1: str, *pathsN) -> str:
    """ Joins N paths and the calls normalized_path() """
    return normalized_path(os.path.join(path1, *pathsN))


def glob_with_extensions(rootdir: str, extensions: List[str]) -> List[str]:
    results = []
    for dirpath, _, dirfiles in os.walk(rootdir):
        for file in dirfiles:
            _, fext = os.path.splitext(file)
            if fext in extensions:
                pathstring = os.path.join(dirpath, file)
                pathstring = normalized_path(pathstring)
                results.append(pathstring)
    return results


def strstr_multi(s: str, substrings: List[str]) -> bool:
    #console(f'file: {s} matches: {substrings}')
    if not substrings: # if no substrings, then match everything
        return True
    for substr in substrings:
        if substr in s:
            return True
    return False


def glob_with_name_match(rootdir: str, pattern_substrings: list, match_dirs=True) -> List[str]:
    results = []
    for dirpath, dirnames, dirfiles in os.walk(rootdir):
        if match_dirs:
            for dir in dirnames:
                if strstr_multi(dir, pattern_substrings):
                    pathstring = os.path.join(dirpath, dir)
                    pathstring = normalized_path(pathstring)
                    results.append(pathstring)
        for file in dirfiles:
            if strstr_multi(file, pattern_substrings):
                pathstring = os.path.join(dirpath, file)
                pathstring = normalized_path(pathstring)
                results.append(pathstring)
    return results


def glob_folders_with_name_match(rootdir: str, pattern_substrings: List[str]):
    results = []
    for dirpath, _, _ in os.walk(rootdir):
        if strstr_multi(dirpath, pattern_substrings):
            results.append(normalized_path(dirpath))
    return results


def is_dir_empty(dir: str): # no files?
    if not os.path.exists(dir): return True
    _, _, filenames = next(os.walk(dir))
    return len(filenames) == 0


def has_tag_changed(old_tag_file: str, new_tag: str):
    if not os.path.exists(old_tag_file):
        return True
    old_tag = read_text_from(old_tag_file)
    if old_tag != new_tag:
        console(f" tagchange '{old_tag.strip()}'\n"+
                f"      ---> '{new_tag.strip()}'")
        return True
    return False


def read_text_from(file_path: str) -> str:
    return pathlib.Path(file_path).read_text()


def write_text_to(file: str, text: str):
    dirname = os.path.dirname(file)
    if not os.path.exists(dirname):
        os.makedirs(dirname, exist_ok=True)
    pathlib.Path(file).write_text(text, encoding='utf-8')


def read_lines_from(file: str) -> List[str]:
    if not os.path.exists(file):
        return []
    with pathlib.Path(file).open(encoding='utf-8') as f:
        return f.readlines()


def get_file_size_str(size):
    """
    Returns file size as a human readable string, eg 96.5KB, or 100.1MB
    """
    if size < 128: return f'{size}B' # only show bytes for really small < 0.1 KB sizes
    if size < (1024*1024): return f'{size/1024:.1f}KB'
    if size < (1024*1024*1024): return f'{size/(1024*1024):.1f}MB'
    return f'{size/(1024*1024):.2}GB'


def get_time_str(seconds: float):
    if seconds < 1: return f'{int(seconds*1000)}ms'
    if seconds < 60: return f'{seconds:.1f}s'
    if seconds < 60*60: return f'{int(seconds%60)}m {int(seconds/60)}s'
    if seconds < 24*60*60: return f'{int(seconds%(60*60))}h {int(seconds%60)}m {int(seconds/60)}s'
    return f'{int(seconds%(24*60*60))}d {int(seconds%(60*60))}h {int(seconds%60)}m {int(seconds/60)}s'


def download_file(remote_url:str, local_dir:str, force=False, message=None):
    local_file = os.path.join(local_dir, os.path.basename(remote_url))
    if not force and os.path.exists(local_file): # download file?
        console(f"    Using locally cached {local_file}")
        return local_file
    start = time.time()
    if not os.path.exists(local_dir):
        os.makedirs(local_dir, exist_ok=True)

    # TODO: this causes issues inside some secure networks
    if remote_url.startswith('https://'):
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_OPTIONAL
    else:
        ctx = None

    with request.urlopen(remote_url, context=ctx, timeout=15) as urlfile:
        size = int(urlfile.info()['Content-Length'].strip())
        if not message: message = f'Downloading {remote_url}'
        print(f'{message} {get_file_size_str(size) if size else "unknown size"}')
        if not size:
            return None

        # for 100MB file, interval = 1
        # for 10MB file, interval = 10
        # for 1MB file, interval = 100 (so essentially disabled)
        report_interval = max(1, int((100*1024*1024) / size))
        transferred = 0
        lastpercent = 0
        print(f'    |{" ":50}<| {0:>3}%', end='')
        with open(local_file, 'wb') as output:
            while transferred < size:
                data = urlfile.read(32*1024) # large chunks plz
                if not data: break
                output.write(data)
                transferred += len(data)
                if report_interval < 100:
                    percent = int((transferred / size) * 100.0)
                    if abs(lastpercent - percent) >= report_interval:
                        lastpercent = percent
                        n = int(percent / 2)
                        right = '=' * n
                        left = ' ' * int(50 - n)
                        elapsed = time.time() - start
                        print(f'\r    |{left}<{right}| {percent:>3}% ({get_time_str(elapsed)})', end='')

    # report actual percent here, just incase something goes wrong
    elapsed = time.time() - start
    percent = int((transferred / size) * 100.0)
    print(f'\r    |<{"="*50}| {percent:>3}% ({get_time_str(elapsed)})')
    return local_file


def unzip(local_zip: str, extract_dir: str, pwd: str = None):
    """
    Attempts to unzip an archive, throws on failure.
    Only extracts the files if their current size or modified time mismatches.
    Always sets modified time from the zipfile info.
    Preserves symlinks. And sets the correct file permission attributes.
    Returns # of files actually extracted.
    """
    def get_zipinfo_datetime(zipmember: zipfile.ZipInfo) -> datetime:
        zt = zipmember.date_time # tuple: year, month, day, hour, min, sec
        # ZIP uses localtime
        return datetime(zt[0], zt[1], zt[2], zt[3], zt[4], zt[5], tzinfo=tz.tzlocal())

    def has_file_changed(zipmember: zipfile.ZipInfo, dst_path):
        st: os.stat_result = None
        try:
            st = os.stat(dst_path, follow_symlinks=False)
            if st.st_size != zipmember.file_size:
                return True
            dst_mtime: datetime = datetime.fromtimestamp(st.st_mtime, tz=tz.tzlocal())
            src_mtime = get_zipinfo_datetime(zipmember)
            if dst_mtime != src_mtime:
                return True
        except (OSError, ValueError):
            return True # does not exist
        return False

    def make_symlink(zipmember: zipfile.ZipInfo, symlink_location, is_directory):
        target = zip.read(zipmember, pwd=pwd).decode('utf-8')
        if os.path.lexists(symlink_location):
            os.remove(symlink_location)
        os.symlink(target, symlink_location, target_is_directory=is_directory)

    unzipped_files: List[Tuple[zipfile.ZipFile, str]] = []

    with zipfile.ZipFile(local_zip, "r") as zip:
        for zipmember in zip.infolist():
            dst_path = os.path.normpath(os.path.join(extract_dir, zipmember.filename))
            mode = zipmember.external_attr >> 16
            is_symlink = stat.S_ISLNK(mode)
            #what = 'DIR' if zipmember.is_dir() else 'FILE'
            #what = what + ' LINK' if is_symlink else what
            #print(f'{what} {zipmember.filename} S_IMODE={stat.S_IMODE(mode):0o} S_IFMT={stat.S_IFMT(mode):0o}')
            if zipmember.is_dir():  # make dirs if needed
                if is_symlink:
                    make_symlink(zipmember, dst_path, is_directory=True)
                else:
                    os.makedirs(dst_path, exist_ok=True)
            elif has_file_changed(zipmember, dst_path):  # only extract if file appears to be modified
                unzipped_files.append((zipmember, dst_path))
                if is_symlink:
                    make_symlink(zipmember, dst_path, is_directory=False)
                else:
                    with zip.open(zipmember, pwd=pwd) as src, open(dst_path, "wb") as dst:
                        shutil.copyfileobj(src, dst)
        for zipmember, dst_path in unzipped_files:
            # set the correct permissions for files and folders
            perm = stat.S_IMODE(zipmember.external_attr >> 16)
            os.chmod(dst_path, perm)
            # always set the modification date from the zipmember timestamp,
            # this way we can avoid unnecessarily modifying files and causing full rebuilds
            time = get_zipinfo_datetime(zipmember)
            #print(f'    | {dst_path} {time}')
            mtime = time.timestamp()
            if System.windows:
                os.utime(dst_path, times=(mtime, mtime))
            else:
                os.utime(dst_path, times=(mtime, mtime), follow_symlinks=False)

    return len(unzipped_files)


def try_unzip(local_file:str, extract_dir:str) -> bool:
    """
    Attempts to unzip an archive, returns a tuple (success: bool, num_extracted: int)
    If (success: True, num_extracted: 0) is returned, it means none of the destination files
    were different from the zip contents, and zero extractions were performed
    """
    try:
        files_extracted = unzip(local_file, extract_dir)
        return (True, files_extracted)
    except zipfile.BadZipFile as e:
        return (False, -1)


def download_and_unzip(remote_file, extract_dir, local_file):
    if local_file and os.path.exists(local_file):
        console(f"Skipping {os.path.basename(remote_file)} because {local_file} exists.")
        return extract_dir
    local_file = download_file(remote_file, extract_dir)
    if not local_file:
        return None
    unzip(local_file, extract_dir)
    console(f'Extracted {local_file} to {extract_dir}')
    return extract_dir


def _should_copy(src: str, dst: str):
    src_stat = None
    try:
        src_stat = os.stat(src)
    except (OSError, ValueError):
        return False # does not exist, nothing to copy

    dst_stat = None
    try:
        dst_stat = os.stat(dst)
    except (OSError, ValueError):
        return True # dst doesn't exist, definitely need to copy it

    if src_stat.st_size != dst_stat.st_size:
        #console(f'_should_copy true src.size != dst.size\n┌──{src}\n└─>{dst}')
        return True
    if src_stat.st_mtime != dst_stat.st_mtime:
        #console(f'_should_copy true src.mtime != dst.mtime\n┌<──{src}\n└──> {dst}')
        return True
    #console(f'skip {dst}')
    return False


def _passes_filter(src_file: str, filter: List[str]) -> bool:
    if not filter: return True
    for f in filter:
        if src_file.endswith(f):
            return True
    return False


def copy_file(src: str, dst: str, filter: List[str]) -> bool:
    """
        Copies a single file if it passes the filter and
        if it has changed, returns TRUE if copied
    """
    if _passes_filter(src, filter) and _should_copy(src, dst):
        #console(f'copy {src}\n --> {dst}')
        shutil.copy2(src, dst)
        return True
    return False


def copy_dir(src_dir: str, out_dir: str, filter=None) -> bool:
    """
        Copies an entire dir if it passes the filter and
        if the individual files have changed.
        Returns TRUE if any files were copied.
    """
    if not os.path.exists(src_dir):
        raise RuntimeError(f'copy_dir: {src_dir} does not exist!')
    copied = False
    root = os.path.dirname(src_dir)
    for fulldir, _, files in os.walk(src_dir):
        reldir = fulldir[len(root):].lstrip('\\/')
        if reldir:
            dst_folder = os.path.join(out_dir, reldir)
            os.makedirs(dst_folder, exist_ok=True)
        else:
            dst_folder = out_dir
        for file in files:
            src_file = os.path.join(fulldir, file)
            dst_file = os.path.join(dst_folder, file)
            copied |= copy_file(src_file, dst_file, filter)
    return copied


def copy_if_needed(src: str, dst: str, filter=None) -> bool:
    """ Copies src -> dst  dir/file  if needed and returns TRUE if anything was copied """
    #console(f'COPY {src} --> {dst}')
    if os.path.isdir(src):
        return copy_dir(src, dst, filter)
    else:
        return copy_file(src, dst, filter)




from __future__ import annotations
import os, ftplib, traceback, getpass
from typing import List, Tuple, TYPE_CHECKING
from urllib.error import HTTPError

from .types.git import Git
from .types.local_source import LocalSource
from .types.artifactory_pkg import ArtifactoryPkg
from .types.dep_source import DepSource
from .types.asset import Asset
from .utils.system import Color, System, console, error
import mama.package as package
from .util import download_file, normalized_join, read_lines_from, try_unzip


if TYPE_CHECKING:
    from .build_target import BuildTarget
    from .build_config import BuildConfig


def artifactory_archive_name(target:BuildTarget):
    """
    Constructs archive name for papa deploy packages in the form of:
    {name}-{platform}-{compiler}-{arch}-{build_type}-{commit_hash}
    Example: opencv-linux-x64-gcc9-release-df76b66
    """
    p:ArtifactoryPkg = target.dep.dep_source

    # LocalSource has no archive name, except for ROOT packages
    if p.is_src and not target.dep.is_root:
        return None

    # if this is an ArtifactoryPkg with full name of the archive
    if p.is_pkg and p.fullname:
        return p.fullname

    # automatically build name of the package
    version = ''
    if target.dep.is_root:
        version = Git.get_current_repository_commit(target.dep)
        if not version:
            return None # nothing to do at this point
    elif p.is_pkg:
        version = p.version
    elif p.is_git:
        git:Git = p
        version = git.get_commit_hash(target.dep)
        if not version:
            return None # nothing to do at this point

    name = target.name
    # triplets information to make this package platform unique
    platform, os_major, os_minor = target.config.get_distro_info()
    compiler = target.config.compiler_version()
    arch = target.config.arch # eg 'x86', 'arm64'
    build_type = 'release' if target.config.release else 'debug'

    return f'{name}-{platform}-{os_major}-{compiler}-{arch}-{build_type}-{version}'


keyr = None
def _get_keyring():
    global keyr
    if not keyr: # lazy init keyring, because it loads certs and other slow stuff
        import keyring
        if System.linux:
            import importlib
            cryptfile = importlib.import_module('keyrings.cryptfile.cryptfile')
            kr = cryptfile.CryptFileKeyring()
            kr.keyring_key = f'mamabuild-{os.getenv("USER")}'
            keyring.set_keyring(kr)
        keyr = keyring
    return keyr


def _get_artifactory_ftp_credentials(config:BuildConfig, url:str):
    # get the values from ENV
    username = os.getenv('MAMA_ARTIFACTORY_USER', None)
    password = os.getenv('MAMA_ARTIFACTORY_PASS', '')  # empty password as default
    if username is not None:
        return username, password

    if config.artifactory_auth == 'store':
        username = _get_keyring().get_password('mamabuild', f'username-{url}')
        password = _get_keyring().get_password('mamabuild', f'password-{url}')
        if username is not None:
            return username, password

    username = input(f'{url} username: ').strip()
    if not username:
        raise EnvironmentError(f'Artifactory user missing: try setting MAMA_ARTIFACTORY_USER env variable.')

    password = getpass.getpass(f'{username}@{url} password: ').strip()
    if password is None: # None on CI
        raise EnvironmentError(f'Artifactory user missing: try setting MAMA_ARTIFACTORY_PASS env variable.')
    return username, password


def _remove_artifactory_ftp_credentials(url:str):
    if _get_keyring().get_password('mamabuild', f'username-{url}'):
        _get_keyring().delete_password('mamabuild', f'username-{url}')
    if _get_keyring().get_password('mamabuild', f'password-{url}'):
        _get_keyring().delete_password('mamabuild', f'password-{url}')


def _store_artifactory_ftp_credentials(config:BuildConfig, url, username, password):
    if config.artifactory_auth == 'store':
        _get_keyring().set_password('mamabuild', f'username-{url}', username)
        _get_keyring().set_password('mamabuild', f'password-{url}', password)


def artifactory_ftp_login(ftp:ftplib.FTP_TLS, config:BuildConfig, url:str):
    connected = False
    while True:
        username, password = _get_artifactory_ftp_credentials(config, url)
        if not connected:
            if config.verbose:
                console(f'  - Artifactory Connect {url}')
            ftp.connect(url, timeout=60)
            connected = True
        try:
            ftp.login(username, password)
            _store_artifactory_ftp_credentials(config, url, username, password)
        except ftplib.Error as e:
            console(f'artifactory login failed: {e}')
            _remove_artifactory_ftp_credentials(url)
        else:
            return # success


def artifactory_sanitize_url(url: str):
    return url.replace('ftp://', '').replace('http://','').replace('https://','')


def artifactory_upload(ftp:ftplib.FTP_TLS, target_name:str, file_path:str):
    size = os.path.getsize(file_path)
    transferred = 0
    lastpercent = 0
    with open(file_path, 'rb') as f:
        def print_progress(bytes):
            nonlocal transferred, lastpercent, size
            transferred += len(bytes)
            percent = int((transferred / size) * 100.0)
            if abs(lastpercent - percent) >= 5:
                lastpercent = percent
                n = int(percent / 2)
                left = '=' * n
                right = ' ' * int(50 - n)
                print(f'\r    |{left}>{right}| {percent:>3} %', end='')
        print(f'    |>{" ":50}| {0:>3} %', end='')
        # chdir into FTP_ROOT/target_name/
        try:
            ftp.cwd(target_name)
        except:
            ftp.mkd(target_name) # create subdirectory if needed
            ftp.cwd(target_name)
        ftp.storbinary(f'STOR {os.path.basename(file_path)}', f, callback=print_progress)
        print(f'\r    |{"="*50}>| 100 %')


def artifact_already_exists(ftp:ftplib.FTP_TLS, file_path:str):
    items = []
    ftp.dir(os.path.basename(file_path), items.append)
    return len(items) > 0 # the file already exists


def artifactory_upload_ftp(target:BuildTarget, file_path:str) -> bool:
    config = target.config
    url = config.artifactory_ftp
    if not url: raise RuntimeError(f'Artifactory Upload failed: artifactory_ftp not set by config.set_artifactory_ftp()')
    if config.verbose: console(f'  - Artifactory Upload {file_path}\n {"":12}-> {url}')

    with ftplib.FTP_TLS() as ftp:
        try:
            # sanitize url for ftplib
            url = artifactory_sanitize_url(url)
            artifactory_ftp_login(ftp, config, url)
            if config.if_needed and artifact_already_exists(ftp, file_path):
                if config.verbose: console(f'  - Artifactory Upload skipped: artifact already exists')
                return False # skip upload
            artifactory_upload(ftp, target.name, file_path)
            return True
        except:
            traceback.print_exc()
        finally:
            ftp.quit()
    return False


def make_dep_source(s:str) -> DepSource:
    if s.startswith('git '): return Git.from_papa_string(s[4:])
    if s.startswith('pkg '): return ArtifactoryPkg.from_papa_string(s[4:])
    if s.startswith('src '): return LocalSource.from_papa_string(s[4:])
    raise RuntimeError(f'Unrecognized dependency source: {s}')


def artifactory_load_target(target:BuildTarget, deploy_path, num_files_copied) -> Tuple[bool, list]:
    """
    Reconfigures `target` from {deployment_path}/papa.txt.
    Returns (fetched:bool, dep_sources:list)
    """
    papa_list = normalized_join(deploy_path, 'papa.txt')
    if not os.path.exists(papa_list):
        error(f'    {target.name}  Artifactory Load failed because {papa_list} does not exist')
        return (False, None)

    if target.config.verbose:
        if num_files_copied != 0:
            console(f'    {target.name}  Artifactory Load ({num_files_copied} files were copied)', color=Color.YELLOW)
        else:
            console(f'    {target.name}  Artifactory Load (no files modified)', color=Color.GREEN)

    project_name = None
    dependencies: list = []
    includes: List[str] = []
    libs: List[str] = []
    syslibs: List[str] = []
    assets: List[Asset] = []

    def append_to(to:list, line):
        to.append(normalized_join(deploy_path, line[2:].strip()))

    for line in read_lines_from(papa_list):
        if   line.startswith('P '): project_name = line[2:].strip()
        elif line.startswith('D '): dependencies.append(make_dep_source(line[2:].strip()))
        elif line.startswith('I '): append_to(includes, line)
        elif line.startswith('L '): append_to(libs, line)
        elif line.startswith('S '): append_to(syslibs, line)
        elif line.startswith('A '):
            relpath = line[2:].strip()
            fullpath = normalized_join(deploy_path, relpath)
            assets.append(Asset(relpath, fullpath, None))

    if project_name != target.name:
        error(f'    {target.name}  Artifactory Load failed because {papa_list} ProjectName={project_name} mismatches!')
        return (False, None)

    target.dep.from_artifactory = True
    target.exported_includes = includes # include folders to export from this target
    target.exported_assets = assets # exported asset files
    package.set_export_libs_and_products(target, libs)
    package.reload_syslibs(target, syslibs) # set exported system libraries

    # for git repos, save the used commit hash status, so that next time the fetch is faster
    if target.dep.dep_source.is_git:
        git: Git = target.dep.dep_source
        git.save_status(target.dep)

    return (True, dependencies)


def _fetch_package(target:BuildTarget, url, archive, cache_dir):
    remote_file = f'http://{url}/{target.name}/{archive}.zip'
    try:
        return download_file(remote_file, cache_dir, force=True, 
                             message=f'    Artifactory fetch {url}/{archive} ')
    except Exception as e:
        if target.config.verbose or target.config.force_artifactory:
            error(f'    Artifactory fetch failed with {e} {url}/{archive}.zip')

        d:DepSource = target.dep.dep_source
        # this is an artifactory pkg, so the url MUST exist
        if d.is_pkg:
            raise RuntimeError(f'Artifactory package {d} did not exist at {url}')

        # if server gives us 404, then we need to wipe the git_status and re-initialize
        # the dependency source from scratch
        if d.is_git:
            d: Git = d
            if isinstance(e, HTTPError) and e.code == 404:
                if target.config.verbose:
                    error(f'    Resetting Git status file: {target.name}')
                d.reset_status(target.dep)

        return None


def unzip_and_load_target(target:BuildTarget, local_file:str) -> Tuple[bool, list]:
    success, num_extracted = try_unzip(local_file, target.dep.build_dir)
    if success:
        return artifactory_load_target(target, target.dep.build_dir, num_files_copied = num_extracted)
    else:
        error(f'    Artifactory unzip failed, possibly corrupt package {local_file}')
        os.remove(local_file) # it's probably corrupted
        return (False, None)


def artifactory_fetch_and_reconfigure(target:BuildTarget) -> Tuple[bool, list]:
    """
    Try to fetch prebuilt package from artifactory
    Returns (fetched:bool, dep_sources:list)
    """
    url = target.config.artifactory_ftp
    if not url:
        return (False, None)

    archive = artifactory_archive_name(target)
    if not archive:
        return (False, None)
    
    cache_dir = target.dep.dep_dir #target.dep.workspace
    local_file = normalized_join(cache_dir, f'{archive}.zip')

    # cache is normally used, however `mama update` will ignore the cache and re-downloads the latest
    if os.path.exists(local_file) and not (target.config.update and target.is_current_target()):
        if (target.is_current_target() or target.config.no_specific_target()) \
            and not target.config.test:
            console(f'    Artifactory cache {local_file}')
        success, deps = unzip_and_load_target(target, local_file)
        if success: return (success, deps)

    # use mama verbose to show the failure msgs
    url = artifactory_sanitize_url(url)
    local_file = _fetch_package(target, url, archive, cache_dir)
    if not local_file:
        return (False, None)
    console(f'    Artifactory unzip {archive}')
    return unzip_and_load_target(target, local_file)




from __future__ import annotations
from typing import List, TYPE_CHECKING
import os, shutil

from .build_dependency import BuildDependency
from .artifactory import artifactory_archive_name, artifactory_upload_ftp
from .util import get_file_size_str, normalized_path, write_text_to, console, copy_if_needed
import mama.package as package

if TYPE_CHECKING:
    from .build_target import BuildTarget


def _gather_dependencies(target:BuildTarget) -> List[BuildDependency]:
    dependecies = []
    for child in target.children():
        dependecies.append(child)
    return dependecies


def _results_contain(results, contains_value):
    for target,value in results:
        if value == contains_value:
            return True
    return False


def _gather(target:BuildTarget, recurse, results:list, get_candidates):
    for value in get_candidates(target):
        if not _results_contain(results, value):
            results.append((target,value))
    if recurse:
        for child in target.children():
            _gather(child.target, True, results, get_candidates)
    return results


def _gather_includes(target:BuildTarget, recurse):
    includes = []
    return _gather(target, recurse, includes, lambda t: t.exported_includes)


def _gather_libs(target:BuildTarget, recurse):
    # gather all libs from the root target
    libs = [(target,l) for l in target.exported_libs]

    # and for children, only gather dynamic libs if recurse is set
    if recurse:
        def get_dylibs(t:BuildTarget):
            for l in t.exported_libs:
                if package.is_a_dynamic_library(l): yield l
        for child in target.children():
            _gather(child.target, recurse, libs, get_dylibs)
    return libs


def _gather_syslibs(target:BuildTarget, recurse):
    syslibs = []
    return _gather(target, recurse, syslibs, lambda t: t.exported_syslibs)


def _gather_assets(target:BuildTarget, recurse):
    assets = []
    return _gather(target, recurse, assets, lambda t: t.exported_assets)


def _append_includes(target:BuildTarget, package_full_path, detail_echo, descr, includes):
    if not includes:
        return # nothing to do
    config = target.config
    includes_root = package_full_path + '/include'
    # TODO: should we include .cpp files for easier debugging?
    includes_filter = ['.h','.hpp','.hxx','.hh','.c','.cpp','.cxx']

    # set the default include
    descr.append(f'I include')

    def append(relpath):
        src_path = abs_include
        dst_dir = includes_root

        # matches default include?
        if relpath == 'include' or relpath == 'include/':
            if detail_echo: console(f'    I ({inctarget.name+")": <16}  include')
            dst_dir = normalized_path(includes_root + '/../')
        else:
            if detail_echo: console(f'    I ({inctarget.name+")": <16}  include/{relpath}')
            descr.append(f'I include/{relpath}')

        if config.verbose: console(f'    copy {src_path}\n      -> {dst_dir}')
        copy_if_needed(src_path, dst_dir, includes_filter)

    relincludes = []  # TODO: what was the point of this again?
    for inctarget, abs_include in includes:
        relpath = os.path.basename(abs_include)
        if not relpath in relincludes:
            relincludes.append(relpath)
            append(relpath)


def papa_deploy_to(target:BuildTarget, package_full_path:str,
                   r_includes:bool, r_dylibs:bool, 
                   r_syslibs:bool, r_assets:bool):
    config = target.config
    detail_echo = config.print and target.is_current_target() and (not config.test)
    if detail_echo: console(f'  - PAPA Deploy {package_full_path}')

    dependencies = _gather_dependencies(target)

    if not os.path.exists(package_full_path): # check to avoid Access Denied errors
        os.makedirs(package_full_path, exist_ok=True)

    # set up project and dependencies
    descr = [ f'P {os.path.basename(package_full_path)}' ]
    for d in dependencies:
        if detail_echo: console(f'    D {d.dep_source}')
        descr.append(f'D {d.dep_source.get_papa_string()}')

    includes = _gather_includes(target, r_includes)
    _append_includes(target, package_full_path, detail_echo, descr, includes)

    libs = _gather_libs(target, r_dylibs)
    for libtarget, lib in libs:
        relpath = package.get_lib_basename(lib) # TODO: how to get a proper relpath??
        descr.append(f'L {relpath}')
        #outpath = os.path.join(package_full_path, relpath)
        outpath = package_full_path
        if detail_echo: console(f'    L ({libtarget.name+")": <16}  {relpath}')
        if config.verbose: console(f'    copy {lib}\n      -> {outpath}')
        copy_if_needed(lib, outpath)

    syslibs = _gather_syslibs(target, r_syslibs)
    for systarget, syslib in syslibs:
        syslib_basename = package.get_lib_basename(syslib)
        descr.append(f'S {syslib_basename}')
        if detail_echo: console(f'    S ({systarget.name+")": <16}  {syslib_basename}')

    assets = _gather_assets(target, r_assets)
    for asstarget, asset in assets:
        descr.append(f'A {asset.outpath}')
        if detail_echo: console(f'    A ({asstarget.name+")": <16}  {asset.outpath}')
        outpath = os.path.join(package_full_path, asset.outpath)

        folder = os.path.dirname(outpath)
        if not os.path.exists(folder):
            os.makedirs(folder, exist_ok=True)
        copy_if_needed(asset.srcpath, outpath)

    write_text_to(os.path.join(package_full_path, 'papa.txt'), '\n'.join(descr))

    # write summary
    if config.print:
        console(f'  PAPA Deployed: {len(includes)} includes, {len(libs)} libs, {len(syslibs)} syslibs, {len(assets)} assets')


def papa_upload_to(target:BuildTarget, package_full_path:str):
    """
    - target: Target which was configured and packaged
    - package_full_path: Full path to deployed PAPA package
    """
    config = target.config
    dst_dir = target.build_dir()
    archive_name = artifactory_archive_name(target)
    if not archive_name:
        raise Exception(f'Could not get archive name for target: {target.name}')

    if config.verbose:
        console(f'    archiving {package_full_path}\n {"":10}-> {dst_dir}/{archive_name}.zip')

    archive = shutil.make_archive(archive_name, 'zip', package_full_path, '.', verbose=True)
    archive_path = dst_dir + '/' + os.path.basename(archive)
    if os.path.exists(archive_path):
        os.remove(archive_path)
    shutil.move(archive, archive_path)

    if config.print:
        size = os.path.getsize(archive_path)
        console(f'  - PAPA Upload {archive_name}  {get_file_size_str(size)}')

    if artifactory_upload_ftp(target, archive_path):
        if config.verbose:
            console(f'  PAPA Uploaded {os.path.basename(archive)}')



import os, shlex, shutil
from signal import SIGTERM
from errno import ECHILD
import subprocess
from time import sleep
from .nonblocking_io import set_nonblocking
from .system import System, console, error


class SubProcess:
    """
    An alternative to subprocess.Popen with redirectable IO
    using fork and forktty on UNIX.

    Windows version uses standard subprocess.Popen with pipes

    Any redirected stdout/stderr which needs to retain its
    terminal colors etc, should use this SubProcess
    """
    def __init__(self, cmd, cwd, env=None, io_func=None):
        self.io_func = io_func
        self.status = None

        env = env if env else os.environ.copy()
        args = shlex.split(cmd)

        executable = args[0]
        if os.path.isfile(executable): # it's something like `./run_tests` or `/usr/bin/gcc`
            executable = os.path.abspath(executable)
        elif System.windows and os.path.isfile(executable + '.exe'):
            executable = os.path.abspath(executable + '.exe')
        else: # lookup from PATH
            executable = shutil.which(args[0])
            if not executable:
                raise OSError(f"SubProcess failed to start: {args[0]} not found in PATH")
        args[0] = executable

        if System.windows:
            self.process = None
            try:
                stdout = subprocess.PIPE if io_func else None
                stderr = subprocess.STDOUT if io_func else None
                self.process = subprocess.Popen(args, cwd=cwd, env=env, shell=True,
                                                universal_newlines=True,
                                                stdout=stdout,
                                                stderr=stderr)
            except Exception as e:
                raise RuntimeError(f"Popen failed {args}: {e}")
        else: # all UNIX based systems support fork or forkpty
            # FD visible only for the parent process, 
            # and can be used to read the child PTY output
            self.parent_fd = 0
            if io_func:
                self.pid, self.parent_fd = os.forkpty()
            else:
                self.pid = os.fork()

            # 0: inside the child process, PID inside the parent process
            if self.pid == -1:
                raise OSError(f"SubProcess failed to start: {cmd}")
            elif self.pid == 0: # child process:
                if cwd: os.chdir(cwd)
                # execve: universal, but requires full path to program
                os.execve(executable, args, env)
            else: # parent process:
                # set the parent FD as non-blocking, otherwise the async tasks will never finish
                set_nonblocking(self.parent_fd)


    def close(self):
        self.kill()
        if System.windows:
            self.process.wait(1.0)
            self.process = None
        else:
            if self.parent_fd:
                os.close(self.parent_fd)
                self.parent_fd = 0


    def kill(self):
        if System.windows:
            self.process.kill()
        else:
            pid, self.pid = (self.pid, 0)
            if pid > 0:
                try:
                    os.kill(pid, SIGTERM)
                except:
                    pass


    def try_wait(self):
        """ Returns EXIT_STATUS int if process has finished, otherwise None """
        if System.windows:
            self.status = self.process.poll()
            return self.status
        else:
            try:
                r, status = os.waitpid(self.pid, os.WNOHANG)
                if r == self.pid: # r == pid: process finished
                    self.status = self._handle_exitstatus(status)
            except OSError as e:
                if e.errno == ECHILD:
                    self.status = -1 # ECHILD: no such child
            return self.status


    def _handle_exitstatus(self, status):
        if os.WIFSIGNALED(status):
            return -os.WTERMSIG(status)
        elif os.WIFEXITED(status):
            return os.WEXITSTATUS(status)
        return -1


    def _parse_lines(self, text: str):
        end = len(text)
        start = 0
        line = ''
        while start < end:
            current = text.find('\n', start)
            if current != -1:
                eol = current
                if (eol-start) > 0 and text[eol-1] == '\r':
                    eol -= 1 # drop the '\r'
                line = text[start:eol]
                start = current + 1
            else: # last token:
                line = text[start:]
                start = end
            self.io_func(line)


    def read_output(self) -> bool:
        """ 
        Returns TRUE if output was read.
        Calls self.io_func(line) for every line that was read.
        Newlines are INCLUDED.
        """
        try:
            if System.windows:
                if not self.process.stdout or self.process.stdout.closed:
                    return False

                text = self.process.stdout.readline()
                # console(f'line: {text} status={self.process.poll()}', end='')
                got_bytes = len(text) > 0
                if self.io_func and got_bytes:
                    self._parse_lines(text)
                return got_bytes
            else:
                if not self.parent_fd:
                    return False

                data: bytes = os.read(self.parent_fd, 8192)
                got_bytes = len(data) > 0
                if self.io_func and got_bytes:
                    text = data.decode()
                    self._parse_lines(text)
                return got_bytes
        except OSError as _:
            # when in non-blocking IO, EAGAIN will be thrown if there's no data
            # and when the other process closes the pipes
            return False


    def read_outputs(self, max_blocks=-1) -> bool:
        """ Reads output multiple times until max_blocks calls of read_output() are done """
        num_reads = 0
        while self.read_output():
            num_reads += 1
            if max_blocks != -1 and num_reads >= max_blocks:
                break # we've read enough
        return num_reads > 0


    @staticmethod
    def run(cmd, cwd=None, env=None, io_func=None):
        """
        Runs the titled sub-process with `cmd` using fork or forktty if io_func is set
        - cmd: full command string
        - cwd: working dir for the subprocess
        - env: execution environment, or None for default env
        - io_func: if set, this callback will receive each line from output
                   if None, then output is echoed as normal to stdout/stderr

        ```
        SubProcess.run('tool', 'cmake xyz', env)
        SubProcess.run('tool', 'cmake xyz', io_func=lambda line: print(line))
        ```
        """
        p = SubProcess(cmd, cwd, env, io_func=io_func)
        try:
            while p.try_wait() is None:
                p.read_outputs(max_blocks=1)
                sleep(0.01)
            p.read_outputs() # read any trailing output
        finally:
            p.close()
        return p.status


def execute(command, echo=False, throw=True):
    """ 
    Executes a command and returns the status code.
    - command: command string
    - echo: if True, prints the command to console
    - throw: if True, throws exception on status_code != 0
    - returns: status code
    """
    if echo: console(command)
    retcode = os.system(command)
    if throw and retcode != 0:
        raise RuntimeError(f'{command} failed with return code {retcode}')
    return retcode


# TODO: use new SubProcess.run instead
def execute_piped(command, cwd=None, timeout=None, throw=True):
    """
    Executes a command and returns the piped outout string
    - command: command string
    - cwd: working dir for the subprocess
    - timeout: timeout in seconds
    - throw: if True, throws exception on status_code != 0
    - returns: output string or None if throw=False
    """
    if not isinstance(command, list):
        command = shlex.split(command)
    try:
        cp = subprocess.run(command, stdout=subprocess.PIPE, cwd=cwd, timeout=timeout)
        return cp.stdout.decode('utf-8').rstrip()
    except Exception as e:
        if throw:
            raise RuntimeError(f'subprocess.Run {command} failed: {e}')
        else:
            return None


def execute_echo(cwd, cmd, exit_on_fail=False):
    """
    Wrapper around SubProcess.run(), by default throws if exit_status != 0
    - cwd: working dir for the subprocess
    - cmd: command string
    - exit_on_fail: if True, exits the application with exit_status
    """
    exit_status = -1
    throw_on_fail = not exit_on_fail
    try:
        exit_status = SubProcess.run(cmd, cwd, io_func=None)
    except:
        error(f'SubProcess exited cwd={cwd} cmd={cmd}')
        if throw_on_fail:
            raise
    if exit_status != 0:
        if throw_on_fail:
            raise RuntimeError(f'Execute {cmd} failed with error: {exit_status}')
        elif exit_on_fail:
            exit(exit_status)


def execute_piped_echo(cwd, cmd, echo=True):
    """
    Wrapper around SubProcess.run(), returns status code with piped output (status, output).
    - cwd: working dir for the subprocess
    - cmd: command string
    - echo: if True, also prints the output to console
    - returns: (exit_status, output_string)
    """
    try:
        exit_status = -1
        output = ''
        def handle_output(line:str):
            nonlocal output
            if echo: print(line)
            output += line
            output += '\n' # newline is not included
        exit_status = SubProcess.run(cmd, cwd, io_func=handle_output)
        return (exit_status, output)
    except Exception as e:
        return (-1, f'{output}{e}')



import os, sys

IS_POSIX = 'linux' in sys.platform.lower() or 'darwin' in sys.platform.lower()

if IS_POSIX:
    def set_nonblocking(fd):
        import fcntl
        flag = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, flag | os.O_NONBLOCK)
else:
    def set_nonblocking(fd):
        import msvcrt
        from ctypes import windll, byref, wintypes, WinError, POINTER
        from ctypes.wintypes import HANDLE, DWORD, BOOL
        PIPE_NOWAIT = DWORD(0x00000001)
        def pipe_no_wait(pipefd):
            SetNamedPipeHandleState = windll.kernel32.SetNamedPipeHandleState
            SetNamedPipeHandleState.argtypes = [HANDLE, POINTER(DWORD), POINTER(DWORD), POINTER(DWORD)]
            SetNamedPipeHandleState.restype = BOOL
            h = msvcrt.get_osfhandle(pipefd)
            res = windll.kernel32.SetNamedPipeHandleState(h, byref(PIPE_NOWAIT), None, None)
            if res == 0:
                print(WinError())
                return False
            return True
        return pipe_no_wait(fd)



import os, runpy, inspect

from .utils.system import console
from .util import path_join, read_text_from, write_text_to

def parse_mamafile(config, target_class, mamafile):
    if not mamafile or not os.path.exists(mamafile):
        return None, None
    #console(f'loaded_mamafile: {mamafile}')

    loaded_globals = runpy.run_path(mamafile)
    for key, value in loaded_globals.items():
        if inspect.isclass(value) and issubclass(value, target_class):
            # print(f'found {key}(BuildTarget): {value}')
            return key, value
    raise RuntimeError(f'No BuildTarget class found in mamafile: {mamafile}')

def update_modification_tag(config, file, tagfile):
    if not os.path.exists(file):
        return False

    filetime = os.path.getmtime(file)
    if not os.path.exists(tagfile):
        os.makedirs(os.path.dirname(tagfile), exist_ok=True)
        if config.verbose: console(f'Update tagfile: {tagfile}')
        write_text_to(tagfile, str(filetime))
        return True

    tagtime = float(read_text_from(tagfile))
    if filetime != tagtime:
        if config.verbose: console(f'Update tagfile: {tagfile}')
        write_text_to(tagfile, str(filetime))
        return True

    if config.verbose: console(f'No Changes {file}')
    return False

## Return: TRUE if mamafile.py was modified
def update_mamafile_tag(config, mamafile, build_dir):
    mamafiletag = path_join(build_dir, 'mamafile_tag')
    return update_modification_tag(config, mamafile, mamafiletag)

## Return: TRUE if CMakeLists.txt was modified
def update_cmakelists_tag(config, cmakelists, build_dir):
    cmakeliststag = path_join(build_dir, 'cmakelists_tag')
    return update_modification_tag(config, cmakelists, cmakeliststag)



import os
from typing import Callable
from mama.utils.system import System, console

class Mips:
    def __init__(self, config):
        self.config = config
        self.toolchain_file = None
        self.supported_arches = ['mips', 'mipsel', 'mips64', 'mips64el']
        self.mips_arch = 'mipsel' # prefer little endian mips
        self.gcc_prefix = ''
        self.libs_path = ''
        self.include_paths = []


    # returns the current path prefix where the compiler can be found
    # example: "/opt/mipsel-openwrt-linux/bin/mipsel-openwrt-linux-"
    def compiler_prefix(self):
        if not self.gcc_prefix: self.init_default()
        return self.gcc_prefix


    def includes(self):
        return self.include_paths


    def init_default(self):
        if not self.gcc_prefix:
            self.init_toolchain(self.mips_arch)


    def init_toolchain(self, arch, toolchain_dir=None, toolchain_file=None):
        if arch not in self.supported_arches:
            raise RuntimeError(f'Unsupported MIPS arch: {arch}')
        if toolchain_file and not os.path.exists(toolchain_file):
            raise FileNotFoundError(f'Toolchain file not found: {toolchain_file}')
        if toolchain_dir and not os.path.exists(toolchain_dir):
            raise FileNotFoundError(f'Toolchain directory not found: {toolchain_dir}')
        if not System.linux:
            raise RuntimeError('MIPS only supported on Linux')

        # direct system installed MIPS toolchain
        self.mips_arch = arch
        self.toolchain_file = toolchain_file # additional toolchain to specify sysroot details

        arch_compiler = f'/usr/bin/{arch}-linux-gnu-gcc'
        if os.path.exists(arch_compiler):
            self.gcc_prefix = f'/usr/bin/{arch}-linux-gnu-'
            libs_path = f'/usr/{arch}-linux-gnu/lib'
            if os.path.exists(libs_path):
                self.libs_path = libs_path
            if self.config.print:
                console(f'Found MIPS tools: {self.gcc_prefix}gcc')
                if libs_path:
                    console(f'  MIPS syslibs: {libs_path}')
            return # success

        raise EnvironmentError('No MIPS toolchain compilers detected, '+
                               f'try "sudo apt-get install g++-{arch}-linux-gnu"')


    def _init_root_path(self, root_path):
        
        pass


    def get_cxx_flags(self, add_flag: Callable[[str,str], None]):
        add_flag('-DMIPS', '1')
        if self.libs_path:
            add_flag(f'-L {self.libs_path}')
        for path in self.includes():
            add_flag(f'-I {path}')

    def get_cmake_build_opts(self) -> list:
        if self.toolchain_file:
            if self.config.print:
                console(f'MIPS Toolchain: {self.toolchain_file}')
            return [
                'MIPS=TRUE',
                f'CMAKE_TOOLCHAIN_FILE="{self.toolchain_file}"'
            ]
        opt = [
            'MIPS=TRUE',
            'CMAKE_SYSTEM_NAME=Linux',
            'CMAKE_SYSTEM_VERSION=1',
            f'CMAKE_SYSTEM_PROCESSOR={self.mips_arch}',
            'CMAKE_FIND_ROOT_PATH_MODE_PROGRAM=ONLY', # Search for compiler tools
            'CMAKE_FIND_ROOT_PATH_MODE_LIBRARY=ONLY', # Search for libraries and headers in the target directories only
            'CMAKE_FIND_ROOT_PATH_MODE_INCLUDE=ONLY',
        ]
        return opt
